@@ -44,11 +44,14 @@ static char *input_file;
 static char *hashname;
 
 /* the extension of the file after pre-processing */
-static char *i_extension;
+static const char *i_extension;
 
 /* the name of the temporary pre-processor file */
 static char *i_tmpfile;
 
+/* the name of the cpp stderr file */
+static char *cpp_stderr;
+
 /* the name of the statistics file */
 char *stats_file = NULL;
 
@@ -71,6 +74,8 @@ static struct {
 	{"CPP", "ii"},
 	{"cxx", "ii"},
 	{"CXX", "ii"},
+	{"c++", "ii"},
+	{"C++", "ii"},
 	{NULL, NULL}};
 
 /*
@@ -78,18 +83,60 @@ static struct {
 */
 static void failed(void)
 {
+	char *e;
+
 	/* delete intermediate pre-processor file if needed */
 	if (i_tmpfile) {
 		unlink(i_tmpfile);
 		free(i_tmpfile);
 		i_tmpfile = NULL;
 	}
+
+	/* delete the cpp stderr file if necessary */
+	if (cpp_stderr) {
+		unlink(cpp_stderr);
+		free(cpp_stderr);
+		cpp_stderr = NULL;
+	}
+
+	/* strip any local args */
+	args_strip(orig_args, "--ccache-");
+
+	if ((e=getenv("CCACHE_PREFIX"))) {
+		char *p = find_executable(e, MYNAME);
+		if (!p) {
+			perror(e);
+			exit(1);
+		}
+		args_add_prefix(orig_args, p);
+	}
+
 	execv(orig_args->argv[0], orig_args->argv);
 	cc_log("execv returned (%s)!\n", strerror(errno));
 	perror(orig_args->argv[0]);
 	exit(1);
 }
 
+
+/* return a string to be used to distinguish temporary files 
+   this also tries to cope with NFS by adding the local hostname 
+*/
+static const char *tmp_string(void)
+{
+	static char *ret;
+
+	if (!ret) {
+		char hostname[200];
+		strcpy(hostname, "unknown");
+		gethostname(hostname, sizeof(hostname)-1);
+		hostname[sizeof(hostname)-1] = 0;
+		asprintf(&ret, "%s.%u", hostname, (unsigned)getpid());
+	}
+
+	return ret;
+}
+
+
 /* run the real compiler and put the result in cache */
 static void to_cache(ARGS *args)
 {
@@ -98,9 +145,9 @@ static void to_cache(ARGS *args)
 	struct stat st1, st2;
 	int status;
 
-	x_asprintf(&tmp_stdout, "%s/tmp.stdout.%d", cache_dir, getpid());
-	x_asprintf(&tmp_stderr, "%s/tmp.stderr.%d", cache_dir, getpid());
-	x_asprintf(&tmp_hashname, "%s/tmp.hash.%d.o", cache_dir, getpid());
+	x_asprintf(&tmp_stdout, "%s/tmp.stdout.%s", cache_dir, tmp_string());
+	x_asprintf(&tmp_stderr, "%s/tmp.stderr.%s", cache_dir, tmp_string());
+	x_asprintf(&tmp_hashname, "%s/tmp.hash.%s.o", cache_dir, tmp_string());
 
 	args_add(args, "-o");
 	args_add(args, tmp_hashname);
@@ -132,7 +179,19 @@ static void to_cache(ARGS *args)
 		if (fd != -1) {
 			if (strcmp(output_file, "/dev/null") == 0 ||
 			    rename(tmp_hashname, output_file) == 0 || errno == ENOENT) {
-				/* we can use a quick method of getting the failed output */
+				if (cpp_stderr) {
+					/* we might have some stderr from cpp */
+					int fd2 = open(cpp_stderr, O_RDONLY);
+					if (fd2 != -1) {
+						copy_fd(fd2, 2);
+						close(fd2);
+						unlink(cpp_stderr);
+						cpp_stderr = NULL;
+					}
+				}
+
+				/* we can use a quick method of
+                                   getting the failed output */
 				copy_fd(fd, 2);
 				close(fd);
 				unlink(tmp_stderr);
@@ -154,7 +213,7 @@ static void to_cache(ARGS *args)
 	    stat(tmp_hashname, &st2) != 0 ||
 	    rename(tmp_hashname, hashname) != 0 ||
 	    rename(tmp_stderr, path_stderr) != 0) {
-		cc_log("failed to rename tmp files\n");
+		cc_log("failed to rename tmp files - %s\n", strerror(errno));
 		stats_update(STATS_ERROR);
 		failed();
 	}
@@ -188,6 +247,12 @@ static void find_hash(ARGS *args)
 
 	hash_start();
 
+	/* when we are doing the unifying tricks we need to include
+           the input file name in the hash to get the warnings right */
+	if (!found_debug) {
+		hash_string(input_file);
+	}
+
 	/* first the arguments */
 	for (i=1;i<args->argc;i++) {
 		/* some arguments don't contribute to the hash. The
@@ -226,17 +291,26 @@ static void find_hash(ARGS *args)
 	/* the compiler driver size and date. This is a simple minded way
 	   to try and detect compiler upgrades. It is not 100% reliable */
 	if (stat(args->argv[0], &st) != 0) {
-		cc_log("Couldn't stat the compiler!?\n");
+		cc_log("Couldn't stat the compiler!? (argv[0]='%s')\n", args->argv[0]);
 		stats_update(STATS_COMPILER);
 		failed();
 	}
 	hash_int(st.st_size);
 	hash_int(st.st_mtime);
 
+	/* possibly hash the current working directory */
+	if (getenv("CCACHE_HASHDIR")) {
+		char *cwd = gnu_getcwd();
+		if (cwd) {
+			hash_string(cwd);
+			free(cwd);
+		}
+	}
+
 	/* now the run */
-	x_asprintf(&path_stdout, "%s/tmp.stdout.%d.%s", cache_dir, getpid(), 
+	x_asprintf(&path_stdout, "%s/tmp.stdout.%s.%s", cache_dir, tmp_string(), 
 		   i_extension);
-	x_asprintf(&path_stderr, "%s/tmp.stderr.%d", cache_dir, getpid());
+	x_asprintf(&path_stderr, "%s/tmp.cpp_stderr.%s", cache_dir, tmp_string());
 
 	args_add(args, "-E");
 	args_add(args, input_file);
@@ -255,19 +329,31 @@ static void find_hash(ARGS *args)
 	   preprocessor output, which means we are sensitive to line number
 	   information. Otherwise we can discard line number info, which makes
 	   us less sensitive to reformatting changes 
+
+	   Note! I have now disabled the unification code by default
+	   as it gives the wrong line numbers for warnings. Pity.
 	*/
-	if (found_debug || getenv("CCACHE_NOUNIFY")) {
+	if (found_debug || !getenv("CCACHE_UNIFY")) {
 		hash_file(path_stdout);
 	} else {
 		if (unify_hash(path_stdout) != 0) {
+			stats_update(STATS_ERROR);
 			failed();
 		}
 	}
 	hash_file(path_stderr);
 
 	i_tmpfile = path_stdout;
-	unlink(path_stderr);
-	free(path_stderr);
+
+	if (!getenv("CCACHE_CPP2")) {
+		/* if we are using the CPP trick then we need to remember this stderr
+		   data and output it just before the main stderr from the compiler
+		   pass */
+		cpp_stderr = path_stderr;
+	} else {	
+		unlink(path_stderr);
+		free(path_stderr);
+	}
 
 	/* we use a N level subdir for the cache path to reduce the impact
 	   on filesystems which are slow for large directories
@@ -300,7 +386,7 @@ static void find_hash(ARGS *args)
    otherwise it returns */
 static void from_cache(int first)
 {
-	int fd_stderr;
+	int fd_stderr, fd_cpp_stderr;
 	char *stderr_file;
 	int ret;
 	struct stat st;
@@ -321,16 +407,24 @@ static void from_cache(int first)
 		return;
 	}
 
+	/* the user might be disabling cache hits */
+	if (first && getenv("CCACHE_RECACHE")) {
+		close(fd_stderr);
+		unlink(stderr_file);
+		free(stderr_file);
+		return;
+	}
+
 	utime(stderr_file, NULL);
 
 	if (strcmp(output_file, "/dev/null") == 0) {
 		ret = 0;
 	} else {
 		unlink(output_file);
-		if (getenv("CCACHE_NOLINK")) {
-			ret = copy_file(hashname, output_file);
-		} else {
+		if (getenv("CCACHE_HARDLINK")) {
 			ret = link(hashname, output_file);
+		} else {
+			ret = copy_file(hashname, output_file);
 		}
 	}
 
@@ -365,6 +459,16 @@ static void from_cache(int first)
 		i_tmpfile = NULL;
 	}
 
+	/* send the cpp stderr, if applicable */
+	fd_cpp_stderr = open(cpp_stderr, O_RDONLY);
+	if (fd_cpp_stderr != -1) {
+		copy_fd(fd_cpp_stderr, 2);
+		close(fd_cpp_stderr);
+		unlink(cpp_stderr);
+		free(cpp_stderr);
+		cpp_stderr = NULL;
+	}
+
 	/* send the stderr */
 	copy_fd(fd_stderr, 2);
 	close(fd_stderr);
@@ -383,21 +487,15 @@ static void from_cache(int first)
 static void find_compiler(int argc, char **argv)
 {
 	char *base;
-	char *path, *tok;
-	struct stat st1, st2;
+	char *path;
 
-	orig_args = args_init();
-	free(orig_args->argv);
-
-	orig_args->argv = argv;
-	orig_args->argc = argc;
+	orig_args = args_init(argc, argv);
 
 	base = basename(argv[0]);
 
 	/* we might be being invoked like "ccache gcc -c foo.c" */
 	if (strcmp(base, MYNAME) == 0) {
-		orig_args->argv++;
-		orig_args->argc--;
+		args_remove_first(orig_args);
 		free(base);
 		if (strchr(argv[1],'/')) {
 			/* a full path was given */
@@ -406,72 +504,36 @@ static void find_compiler(int argc, char **argv)
 		base = basename(argv[1]);
 	}
 
-	path = getenv("CCACHE_PATH");
-	if (!path) {
-		path = getenv("PATH");
-	}
-	if (!path) {
-		cc_log("no PATH variable!?\n");
-		failed();
+	/* support user override of the compiler */
+	if ((path=getenv("CCACHE_CC"))) {
+		base = strdup(path);
 	}
 
-	path = x_strdup(path);
-	
-	/* search the path looking for the first compiler of the right name
-	   that isn't us */
-	for (tok=strtok(path,":"); tok; tok = strtok(NULL, ":")) {
-		char *fname;
-		x_asprintf(&fname, "%s/%s", tok, base);
-		/* look for a normal executable file */
-		if (access(fname, X_OK) == 0 &&
-		    lstat(fname, &st1) == 0 &&
-		    stat(fname, &st2) == 0 &&
-		    S_ISREG(st2.st_mode)) {
-			/* if its a symlink then ensure it doesn't
-                           point at something called "ccache" */
-			if (S_ISLNK(st1.st_mode)) {
-				char *buf = x_realpath(fname);
-				if (buf) {
-					char *p = basename(buf);
-					if (strcmp(p, MYNAME) == 0) {
-						/* its a link to "ccache" ! */
-						free(p);
-						free(buf);
-						continue;
-					}
-					free(buf);
-					free(p);
-				}
-			}
-
-
-			/* found it! */
-			free(path);
-			orig_args->argv[0] = fname;
-			free(base);
-			return;
-		}
-		free(fname);
-	}
+	orig_args->argv[0] = find_executable(base, MYNAME);
 
 	/* can't find the compiler! */
-	perror(base);
-	exit(1);
+	if (!orig_args->argv[0]) {
+		stats_update(STATS_COMPILER);
+		perror(base);
+		exit(1);
+	}
 }
 
 
 /* check a filename for C/C++ extension. Return the pre-processor
    extension */
-static char *check_extension(const char *fname)
+static const char *check_extension(const char *fname)
 {
 	int i;
-	char *p;
+	const char *p;
 
 	p = strrchr(fname, '.');
 	if (!p) return NULL;
 	p++;
 	for (i=0; extensions[i].extension; i++) {
 		if (strcmp(p, extensions[i].extension) == 0) {
+			p = getenv("CCACHE_EXTENSION");
+			if (p) return p;
 			return extensions[i].i_extension;
 		}
 	}
@@ -489,18 +551,29 @@ static void process_args(int argc, char **argv)
 	int found_c_opt = 0;
 	int found_S_opt = 0;
 	struct stat st;
+	char *e;
 
-	stripped_args = args_init();
+	stripped_args = args_init(0, NULL);
 
 	args_add(stripped_args, argv[0]);
 
 	for (i=1; i<argc; i++) {
 		/* some options will never work ... */
-		if (strncmp(argv[i], "-E", 2) == 0 ||
-		    strncmp(argv[i], "-M", 2) == 0) {
+		if (strcmp(argv[i], "-E") == 0) {
 			failed();
 		}
 
+		/* these are too hard */
+		if (strcmp(argv[i], "-fprofile-arcs")==0 ||
+		    strcmp(argv[i], "-fbranch-probabilities")==0 ||
+		    strcmp(argv[i], "-M") == 0 ||
+		    strcmp(argv[i], "-x") == 0) {
+			cc_log("argument %s is unsupported\n", argv[i]);
+			stats_update(STATS_UNSUPPORTED);
+			failed();
+			continue;
+		}
+
 		/* we must have -c */
 		if (strcmp(argv[i], "-c") == 0) {
 			args_add(stripped_args, argv[i]);
@@ -544,24 +617,43 @@ static void process_args(int argc, char **argv)
 			continue;
 		}
 
-		/* options that take an argument */
-		if (strcmp(argv[i], "-I") == 0 ||
-		    strcmp(argv[i], "-include") == 0 ||
-		    strcmp(argv[i], "-L") == 0 ||
-		    strcmp(argv[i], "-D") == 0 ||
-		    strcmp(argv[i], "-isystem") == 0) {
-			if (i == argc-1) {
-				cc_log("missing argument to %s\n", argv[i]);
-				stats_update(STATS_ARGS);
+		/* The user knows best: just swallow the next arg */
+		if (strcmp(argv[i], "--ccache-skip") == 0) {
+			i++;
+			if (i == argc) {
 				failed();
 			}
-						
 			args_add(stripped_args, argv[i]);
-			args_add(stripped_args, argv[i+1]);
-			i++;
 			continue;
 		}
 
+		/* options that take an argument */
+		{
+			const char *opts[] = {"-I", "-include", "-imacros", "-iprefix",
+					      "-iwithprefix", "-iwithprefixbefore",
+					      "-L", "-D", "-U", "-x", "-MF", 
+					      "-MT", "-MQ", "-isystem", "-aux-info",
+					      "--param", "-A", "-Xlinker", "-u",
+					      NULL};
+			int j;
+			for (j=0;opts[j];j++) {
+				if (strcmp(argv[i], opts[j]) == 0) {
+					if (i == argc-1) {
+						cc_log("missing argument to %s\n", 
+						       argv[i]);
+						stats_update(STATS_ARGS);
+						failed();
+					}
+						
+					args_add(stripped_args, argv[i]);
+					args_add(stripped_args, argv[i+1]);
+					i++;
+					break;
+				}
+			}
+			if (opts[j]) continue;
+		}
+
 		/* other options */
 		if (argv[i][0] == '-') {
 			args_add(stripped_args, argv[i]);
@@ -623,6 +715,13 @@ static void process_args(int argc, char **argv)
 		failed();
 	}
 
+
+	/* don't try to second guess the compilers heuristics for stdout handling */
+	if (output_file && strcmp(output_file, "-") == 0) {
+		stats_update(STATS_OUTSTDOUT);
+		failed();
+	}
+
 	if (!output_file) {
 		char *p;
 		output_file = x_strdup(input_file);
@@ -637,10 +736,6 @@ static void process_args(int argc, char **argv)
 		}
 		p[1] = found_S_opt ? 's' : 'o';
 		p[2] = 0;
-#if 0
-		cc_log("Formed output file %s from input_file %s\n", 
-		       output_file, input_file);
-#endif
 	}
 
 	/* cope with -o /dev/null */
@@ -649,6 +744,15 @@ static void process_args(int argc, char **argv)
 		stats_update(STATS_DEVICE);
 		failed();
 	}
+
+	if ((e=getenv("CCACHE_PREFIX"))) {
+		char *p = find_executable(e, MYNAME);
+		if (!p) {
+			perror(e);
+			exit(1);
+		}
+		args_add_prefix(stripped_args, p);
+	}
 }
 
 /* the main ccache driver function */
@@ -672,7 +776,7 @@ static void ccache(int argc, char *argv[])
 	/* if we can return from cache at this point then do */
 	from_cache(1);
 	
-	/* run real compiler, semding output to cache */
+	/* run real compiler, sending output to cache */
 	to_cache(stripped_args);
 
 	/* return from cache */
@@ -699,6 +803,7 @@ static void usage(void)
 	printf("-s                      show statistics summary\n");
 	printf("-z                      zero statistics\n");
 	printf("-c                      run a cache cleanup\n");
+	printf("-C                      clear the cache completely\n");
 	printf("-F <maxfiles>           set maximum files in cache\n");
 	printf("-M <maxsize>            set maximum size of cache (use G, M or K)\n");
 	printf("-h                      this help page\n");
@@ -712,7 +817,7 @@ static int ccache_main(int argc, char *argv[])
 	int c;
 	size_t v;
 
-	while ((c = getopt(argc, argv, "hszcF:M:V")) != -1) {
+	while ((c = getopt(argc, argv, "hszcCF:M:V")) != -1) {
 		switch (c) {
 		case 'V':
 			printf("ccache version %s\n", CCACHE_VERSION);
@@ -730,7 +835,12 @@ static int ccache_main(int argc, char *argv[])
 
 		case 'c':
 			cleanup_all(cache_dir);
-			printf("Cleaned cached\n");
+			printf("Cleaned cache\n");
+			break;
+
+		case 'C':
+			wipe_all(cache_dir);
+			printf("Cleared cache\n");
 			break;
 
 		case 'z':
